use std::io::Write;
use std::{iter, slice};

use bincode::config::Configuration;
use bincode::error::{DecodeError, EncodeError};
use bincode::{Decode, Encode};

const COLUMN_USERNAME_SIZE: usize = 32;
const COLUMN_EMAIL_SIZE: usize = 255;

#[derive(Encode, Decode)]
pub struct Row {
    pub id: u32,
    pub username: [u8; COLUMN_USERNAME_SIZE],
    pub email: [u8; COLUMN_EMAIL_SIZE],
}

pub trait Serialize {
    fn serialize(&self) -> Result<Vec<u8>, EncodeError>;
    fn deserialize(buffer: &[u8]) -> Result<Self, DecodeError>
    where
        Self: Sized;
}

impl Serialize for Row {
    fn serialize(&self) -> Result<Vec<u8>, EncodeError> {
        bincode::encode_to_vec(self, bincode::config::standard())
    }

    fn deserialize(buffer: &[u8]) -> Result<Self, DecodeError> {
        match bincode::decode_from_slice::<Self, Configuration>(buffer, bincode::config::standard())
        {
            Ok((row, _)) => Ok(row),
            Err(e) => Err(e),
        }
    }
}

impl Row {
    pub fn new(id: u32, username: String, email: String) -> Result<Self, String> {
        let username_bytes = username.as_bytes();
        let email_bytes = email.as_bytes();
        if username_bytes.len() > COLUMN_USERNAME_SIZE {
            Err(String::from("username is too long"))
        } else if email_bytes.len() > COLUMN_EMAIL_SIZE {
            Err(String::from("email is too long"))
        } else {
            let username_buffer = &mut [0; COLUMN_USERNAME_SIZE];
            if let Err(_) = pad_buffer(username_buffer, username_bytes) {
                panic!("could not pad username buffer")
            }
            let email_buffer = &mut [0; COLUMN_EMAIL_SIZE];
            if let Err(_) = pad_buffer(email_buffer, username_bytes) {
                panic!("could not pad username buffer")
            }
            Ok(Self {
                id,
                username: *username_buffer,
                email: *email_buffer,
            })
        }
    }

    pub fn print(&self) -> String {
        format!(
            "({}, {}, {})",
            self.id,
            String::from_utf8(self.email.to_vec()).unwrap(),
            String::from_utf8(self.username.to_vec()).unwrap()
        )
    }
}

fn pad_buffer(mut bytes: &mut [u8], source: &[u8]) -> std::io::Result<usize> {
    bytes.write(source)
}

pub struct RowIter<'a> {
    inner: slice::Iter<'a, Row>,
}

impl<'a> iter::Iterator for RowIter<'a> {
    type Item = &'a Row;

    fn next(&mut self) -> Option<Self::Item> {
        self.inner.next()
    }
}
